/*
Arduino for trailer, unit 1

 Use the slave protocol for the homestead arduinos, driven by a host serial input,
 with wired-or output from each UNO.
 
 Command format:
 First byte is unit address (we're using all ascii so far, plenty for this for now)
 Second byte is command
 Further bytes if required for command data, up to 32 total bytes including
 '\n' at end which terminates the command sequence.

 Minor error detection/ignore implemented.  Might need more...
 Protocol stategy is just to have host try again if it doesn't get an answer.
 


Arduino board pins:
A0-A3 are used.  Water level detection

first counter is pin: 5 (hardware T1)
Second counter is pin 2  which is interrupt counting 
Both counters are pulled high and count on falling edge

First DHT22 is pin 4
Second is pin 7
third is pin 6

led pin 9

thermocouple is spi on 10-13

water inlet is pin A4 (SDA)
water dump is pin A5 (SCL)
 */
   /***********************************************************************************
ard commands for unit 1

a is send ack
l led on
o led off
r report
! reset

i let water in
d dump water
c close valves
*/ 
 
// for thermocouple breakout
#include <SPI.h>
#include "Adafruit_MAX31855.h"

// for DHT 22 temp/relative humidity sensors 
#include "DHT.h"
#define NUM_DHTS 3    // how many temp/humid sensors we have
#define DHTTYPE DHT22   // DHT 22  (AM2302)
 
// Hardware pin defines
// give it a name:
#define BAUD_RATE 115200 // pick yer poison  uses pins 0 and 1
// pin 2 is int0 (software counter two)
// pin 3 avaliable
#define DHT1PIN 4     // what pin we're connected to
// pin 5 is hardware counter1, here called counter 1
#define DHT2PIN 6
#define DHT3PIN 7
// pin 8 available
#define LED_PIN 9  // gosh, save a byte
//#define DO   12  // in use, but define not required, I'm using hardware spi
#define CS   10    // for thermocouple amp/breakout board
// probably can't use pin 11 for other things when hardware SPI is used - would have been SPI-out

// analong pins 0-3 used
#define ADCHNLS 4  // no ; after defines in arduino land
#define ADTHRESH 512 // threshold for water detection (probably more than 5 IRL)
// a4, a5 (i2c) available```````````````````````````````````````````````````````````````````````Jvarsyncwiki

#define WATER_IN_PIN A4
#define WATER_DUMP_PIN A5

// stuff for command state machine below

#define MY_ADDRESS '1'  // could be any byte, but we're using all-ascii for now, and '\n' for end of command
#define CMD_BUF_SIZE 32 // in theory, two of these would fit in the uno buffer at once.  Probably don't need this much IRL
#define MAX_TIME 20 // milliseconds to timeout if incomplete command (3ms would do for 115200 and fast commands)
byte Command[CMD_BUF_SIZE]; // command buffer for incoming commands
byte CommandIndex = 0; // where am I in above buffer
byte CommandState = 0; // state variable
unsigned long CommandTime; // time at start of command recieve
////// done with command-slave state machine stuff


/// general globals
#define RATE 2500    // in milliseconds, "sample rate for sensors"

byte counter = 0; // for debug
byte i; // shared iterator, save those bytes...
unsigned long old_millis = 0;   // Z-1 for timing, rather than just delaying

int AD[ADCHNLS];        // values read from the a/d
const byte levels[ADCHNLS] = {56,112,169,225}; // where the probes are in the tank
byte gallons;  // what we report as cistern level

// for DHT temp/humidity sensors
DHT dht1(DHT1PIN, DHTTYPE); // temp, rel humid
DHT dht2(DHT2PIN, DHTTYPE);
DHT dht3(DHT3PIN, DHTTYPE);

float temps[NUM_DHTS]; // place for DHT temperatures
float relhumids[NUM_DHTS]; // place for DHT humidities

// thermocouple breakout board
Adafruit_MAX31855 thermocouple(CS); // note, MAX31855 lib changed to use CLOCK_DIV32 (was 4)

double TcCj;
double TcTmp;

// counters, used to watch water flow meters
volatile unsigned long c1counts;    // total counts from (hardware) counter 1,
// top half incremented by timer overflow
volatile unsigned long c2counts; // just count interrupts from a pin (ext int 0)




//////////////////////////////////////////////// subs called from setup
void tset() // setup timer1 as a counter that interrupts on overflow
{
  c1counts = c2counts = 0; // start fresh
  TCCR1A = 0; // turn off all the weird compare/capture modes etc
  TCCR1B = 6; // clock on falling edge (7 for rising)
  TIFR1 = _BV(TOV1); // clear initial overflow - shouldn't need to, but do.
  TCNT1 = 0; //zero out timer
  TIMSK1 = _BV(TOIE1);// sets the timer overflow interrupt enable bit
  }
/////////////////////////////////////////////////////////
//////// manual isr setup, you bet, saves a whole library
// actual isr for hardware counter 1
ISR(TIMER1_OVF_vect)
{
 c1counts += 0x00010000; // add one to top half of longword
} 
////////////////////////////////////////////////
void intCounter ()
{
 c2counts++; // pretty simple... 
}
////////////////////////////////////////////////
void readCounts()
{ // now gets hardware counter
/* Disable interrupts */
noInterrupts();
/* Read TCNT1 into c1counts low part */
c1counts &= 0xffff0000;
c1counts |= TCNT1; // from the hardware register @@@ there's a better capture register
// c2 stuff not needed due to no differencing now
interrupts(); 
}
////////////////////////////////////////////////
void clearcounts()
{
 TCNT1 = 0; //zero out timer
 c1counts = c2counts = 0;  
}
////////////////////////////////////////////////
void readAD()
{
   gallons = 0; // water level
   for (i=0; i < ADCHNLS; i++)
   {
    AD[i]  = analogRead(i); // delay required?
    if (AD[i] > ADTHRESH) gallons = levels[i];
   }// for each channel
}
////////////////////////////////////////////////
void readTC()
{  // get these in farenheit
  TcCj = thermocouple.readInternal() * 1.8 + 32;
  TcTmp = thermocouple.readCelsius() * 1.8 + 32; 
}
//////////////////////////////////////////////// 
void soft_reset () // software reset, actually a hard RST
{ // so we can reset timers and so on under host control
 asm volatile ("  jmp 0"); 
} // this really is a hard reset, but coming from software

////////////////////////////////////////////////
void readDHTs()
{ // add smoothing later
 relhumids[0] = dht1.readHumidity(); // basement
 temps[0] = dht1.readTemperature(true);
 relhumids[1] = dht2.readHumidity(); // Outdoor
 temps[1] = dht2.readTemperature(true);
 relhumids[2] = dht3.readHumidity(); // indoor
 temps[2] = dht3.readTemperature(true);
}
////////////////////////////////////////////////
void Report()
{ // send whatever we have back to host over serial
 Serial.print(gallons);
/* 
 for (i=0;i<ADCHNLS;i++)
 {
  Serial.print("\t");
  Serial.print (AD[i]);
 }
*/ 
 
 Serial.print("\t");
 Serial.print(c1counts); // may want to turn off interrupts for counters and reset them
 Serial.print("\t");
 Serial.print(c2counts);
 clearcounts(); // ready for next reporting interval
 for (i=0;i<NUM_DHTS;i++)
 {
  Serial.print ("\t");
  Serial.print(temps[i]);
  Serial.print("\t");
  Serial.print(relhumids[i]); 
 }
 Serial.print("\t");
 Serial.print(TcCj);
 Serial.print("\t");
 Serial.print(TcTmp);
 Serial.print("\t");
 if (digitalRead(WATER_IN_PIN) == LOW) {Serial.print ("0");}
 else {Serial.print("1");}
 Serial.print("\t");
 if (digitalRead(WATER_DUMP_PIN) == LOW) {Serial.print ("0");}
 else {Serial.print("1");}
 Serial.println(); 
}

////////////////////////////////////////////////
void CommandMachine()
{
  switch (CommandState)
 {
  case 0: // looking for work to do
  if (Serial.available())
  {
   CommandTime = millis(); // for later possible timeout
   CommandState = 1; // get bytes till done or timeout
  }
  break;

  case 1: // getting command
  while (Serial.available())
  {
   Command[CommandIndex] = Serial.read(); // get a byte if available
//   Serial.print ("received:"); // debug
//   Serial.println(Command[CommandIndex]); // debug
   
   if (Command[CommandIndex] == '\n') // single quotes, else compiler thinks pointer to string
    { // got a full command, go parse it
     CommandState = 2;
     return; // get it on next roundy round of loop() 
    }
   CommandIndex++; // bump the position
   if ((CommandIndex >= CMD_BUF_SIZE) || ((CommandTime + MAX_TIME) < millis())) // enough parens?
   { // error of some kind, reset state machine
    CommandIndex = 0;
    CommandState = 0; // ignore what might have been garbage
//    Serial.println ("Command error"); // debug
   }
  } // end while
  break;
  
  case 2: // parse command
   if (Command[0] != MY_ADDRESS) // is this even for us?
   { // no, ditch it
    CommandIndex = 0;
    CommandState = 0;
    return; // forget it ever happened 
   }
   // It might have been for us, any legality tests might go after this comment.
   // By convention, address of unit is first byte, command is next, data might follow, should be enough.
   // For time consuming or big stuff, use a subroutine call instead of inlining it like this example.
   // Commands are all-ascii for convienience of host (and we slaves).
   switch (Command[1])
    {
     case 'a': // host requests an ack
        Serial.println("ack");  // should work
     break;
     case 'c': // close all valves
     digitalWrite(WATER_IN_PIN, LOW);
     digitalWrite(WATER_DUMP_PIN, LOW);
     break;
     case 'd':
      digitalWrite(WATER_DUMP_PIN, HIGH);
     break;
     case 'i': // let water in
      digitalWrite(WATER_IN_PIN,HIGH);
     break;
     case 'l': // for test, led on
      digitalWrite(LED_PIN, HIGH);   // turn the LED on (HIGH is the voltage level)
     break; // end 'l' command
     case 'o': // led off command
      digitalWrite(LED_PIN, LOW);    // turn the LED off by making the voltage LOW
     break; // end off

     case 'r': // report our "stuff"
      Report();
     break;
     case '!':
      soft_reset(); // allows "hard" reset from host machine
     break;
    }    
  // we are done with this one   
     CommandIndex = 0;
     CommandState = 0;
//     Serial.println("end state 2"); // debug
    
   break; // end state 2
 } // end switch state machine
} // end command state machine routine

////////////////////////////////////////////////////////////////////////
// the setup routine runs once when you press reset:
void setup() {                
  // initialize the digital pin as an output.
  pinMode(LED_PIN, OUTPUT);
  pinMode (WATER_IN_PIN, OUTPUT);
  digitalWrite(WATER_IN_PIN, LOW);
  pinMode (WATER_DUMP_PIN, OUTPUT);
  digitalWrite(WATER_DUMP_PIN, LOW); 
  Serial.begin(BAUD_RATE); // someplace to get commands from
  digitalWrite(5, HIGH); // turns on pullup resistor for first counter
  digitalWrite(2,HIGH); // same as above for the software/interrupt counter

  tset(); // setup/clear timers  after ensuring high input
  attachInterrupt(0,intCounter,FALLING); // works but eats more bytes
  clearcounts(); // make all zero at start

  dht1.begin(); // haven't worked out array of objects yet...
  dht2.begin();
  dht3.begin();
}
///////////////////////////////////////////////////////////////////////
// the loop routine runs over and over again forever:
void loop() {
// other stuff you're doing goes here (data aq & control?)
  CommandMachine(); // look for work to do from host

  if (millis() - old_millis >= RATE) // don't waste (many) cycles waiting for sample time
  { // sample sensors for later response to host
   digitalWrite(LED_PIN, HIGH);   // turn the LED on (HIGH is the voltage level)
   old_millis += RATE; // @@@ reset "fast" time-base
   readDHTs(); // acquire temp/humidity from DHT 22 sensor array
   readCounts();
   readAD();
   readTC();
   digitalWrite(LED_PIN, LOW);   // turn the LED off
  } // end sampling
}
